from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from algorithms.utils.types import SpielGame, SpielState, SpielAction, Policy
    from algorithms.utils.params import Params
    from typing import Tuple, Any


class ZeroBot(ABC):
    """
    Abstract class used by all bot agents.
    Args:
        game: `SpielGame`, a pyspiel game object
        params: `GameParams`, a named tuple with game parameter information
        verbose: `bool`, used for verbose printing
    """
    def __init__(self, game: SpielGame, params: Params, verbose: bool) -> None:
        self._game = game
        self._extractor = params.extractor
        self._verbose = verbose

    @abstractmethod
    def step(self, state: SpielState) -> SpielAction:
        """
        Used in game simulations to return the action the agent would make in a given state.
        Args:
            state: `SpielState`, a pyspiel state object
        Returns:
            an `int` representing the action taken
        """
        raise NotImplementedError

    @abstractmethod
    def action_and_policy(self, state: SpielState) -> Tuple[SpielAction, Policy]:
        """
        Used in game simulations to return both the action the agent would take and its policy for choosing the action.
        Args:
            state: `SpielState`, a pyspiel state object
        Returns:
            a `Tuple[Action, Policy]` with integer representation of the action and numpy representation of the policy
        """
        raise NotImplementedError

    @abstractmethod
    def search(self, state: SpielState) -> Any:
        raise NotImplementedError
