from __future__ import annotations
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf, numpy as np
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from algorithms.utils.types import ModelWeights, TrainingBatch, AdamOptimizer
    from algorithms.utils.params import Params


class Evaluator(ABC):
    def __init__(self, params: Params, worker_id: int = 0) -> None:

        """
        Abstract class for handling network operations such as activations and updating.
        Args:
            params: `Params`, a named tuple with game parameter information
        """
        self._num_actions = params.num_actions
        self._num_players = params.num_players
        self._l2_regularization = params.l2_regularization
        self._extractor = params.extractor

        if params.device == "gpu":
            if not tf.test.is_gpu_available():
                raise ValueError("GPU support is unavailable.")
            gpu_num = worker_id % 4
            gpu_str = 'gpu:{}'.format(gpu_num)
            self._device = tf.device(gpu_str)
        elif params.device == "cpu":
            self._device = tf.device("cpu:0")
        else:
            self._device = params.device

    @abstractmethod
    def get_weights(self) -> ModelWeights:
        """
        Used for returning the weights of all models used
        Returns:
             a variable-length tuple `ModelWeights` where each element is a list of model weight numpy arrays.
        """
        raise NotImplementedError

    @abstractmethod
    def set_weights(self, weights: ModelWeights) -> None:
        """
        Used for setting the weights for all models in the evaluator.
        Args:
            weights: `ModelWeights`, a variable-length tuple, each element is a list of model weight numpy arrays
        """
        raise NotImplementedError

    @abstractmethod
    def update(self, training_examples: TrainingBatch, optimizer: AdamOptimizer, k: int = 6):
        """
        Using training examples unrolled from game histories in the replay buffer, updates the model weights
        """
        raise NotImplementedError

    @abstractmethod
    def update_with_grads(self, optimizer: AdamOptimizer, grads: List[np.array]):
        """
        Using training examples unrolled from game histories in the replay buffer, updates the model weights
        """
        raise NotImplementedError

