// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef THIRD_PARTY_OPEN_SPIEL_GAMES_NANNON_H_
#define THIRD_PARTY_OPEN_SPIEL_GAMES_NANNON_H_

#include <array>
#include <memory>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include "minispiel/spiel.h"

// An implementation of the classic: https://en.wikipedia.org/wiki/Backgammon
// using rule set from
// http://usbgf.org/learn-backgammon/backgammon-rules-and-terms/rules-of-backgammon/
// where red -> 'x' (player 0) and white -> 'o' (player 1).
//
// Currently does not support the doubling cube
//
// Parameters:
//   "scoring_type"    string  Type of scoring for the game: "winloss_scoring"
//                             (default), "enable_gammons", or "full_scoring"

namespace open_spiel::nannon {

inline constexpr const int kNumPlayers = 2;
inline constexpr const int kXPlayerId = 0;
inline constexpr const int kOPlayerId = 1;
inline constexpr const int kPassPos = -1;
inline constexpr const int kDefaultNumPoints = 6;
inline constexpr const int kDefaultNumCheckersPerPlayer = 3;
inline constexpr const int kDefaultNumChanceOutcomes = 6;
inline constexpr const int kMaxGameLength = 1000;

struct CheckerMove {
    // Pass is encoded as (pos, num, hit) = (-1, -1, false).
    int pos;  // 0-3  (0-2 for locations on the board and kBarPos)
    int num;  // 1-6
    bool hit;
    CheckerMove(int _pos, int _num, bool _hit)
        : pos(_pos), num(_num), hit(_hit) {}
    bool operator<(const CheckerMove& rhs) const {
        return (pos * 6 + (num - 1)) < (rhs.pos * 6 + rhs.num - 1);
    }
};
// This is a small helper to track historical turn info not stored in the moves.
// It is only needed for proper implementation of Undo.
struct TurnHistoryInfo {
    int player;
    int prev_player;
    std::vector<int> dice;
    Action action;
    bool move_hit;
    TurnHistoryInfo(int _player, int _prev_player, std::vector<int> _dice, int _action, bool fmh):
        player(_player),
        prev_player(_prev_player),
        dice(std::move(_dice)),
        action(_action),
        move_hit(fmh){}
};

class NannonGame;

class NannonState : public State {
    public:
        NannonState(const NannonState&) = default;
        NannonState(std::shared_ptr<const Game>,
                int n_points,
                int n_chex_per_player,
                int n_die,
                std::vector<std::pair<Action, double>> chance_outcomes,
                std::vector<int> chance_outcome_values,
                int state_encoding_size,
                int home_pos,
                int score_pos);

        Player CurrentPlayer() const override;
        void UndoAction(Player player, Action action) override;
        std::vector<Action> LegalActions() const override;
        std::string ActionToString(Player player, Action move_id) const override;
        std::vector<std::pair<Action, double>> ChanceOutcomes() const override;
        std::string ToString() const override;
        bool IsTerminal() const override;
        std::vector<double> Returns() const override;
        std::string ObservationString(Player player) const override;
        void ObservationTensor(Player player, std::vector<double>* values) const override;
        std::unique_ptr<State> Clone() const override;
        CheckerMove SpielMoveToCheckerMove(Player player, Action action) const;

        // Setter function used for debugging and tests. Note: this does not set the
        // historical information properly, so Undo likely will not work on states
        // set this way!
        void SetState(int cur_player, const std::vector<int>& dice,
                      const std::vector<int>& home, const std::vector<int>& scores,
                      const std::vector<std::vector<int>>& board,
                      const int turns);

        // Returns the opponent of the specified player.
        int Opponent(int player) const;

        int score(int player) const { return scores_[player]; }
        std::vector<int> board(int player) const { return board_[player]; }
        int home(int player) const { return home_[player]; }
        int dice(int i) const { return dice_[i]; }

        // Action encoding / decoding functions.
        Action CheckerMoveToSpielMove(const CheckerMove& move) const;

    protected:
        void DoApplyAction(Action move_id) override;

    private:
        void RollDice(int outcome);
        int PositionFromHome(int player, int spaces) const;
        int PositionFrom(int player, int pos, int spaces) const;
        int CountTotalCheckers(int player) const;
        Action EncodedPassMove() const;
        Action EncodedHomeMove() const;

        bool ApplyCheckerMove(int player, const CheckerMove& move);
        void UndoCheckerMove(int player, const CheckerMove& move);
        std::set<CheckerMove> LegalCheckerMoves(int player) const;

        const int n_points_;
        const int n_chex_per_player_;
        const int n_die_;
        const std::vector<std::pair<Action, double>> chance_outcomes_;
        const std::vector<int> chance_outcome_values_;
        const int state_encoding_size_;
        const int home_pos_;
        const int score_pos_;

        Player cur_player_;
        Player prev_player_;
        Player true_prev_player_;
        int turns_;
        int x_turns_;
        int o_turns_;
        std::vector<int> dice_;
        std::vector<int> scores_;  // Checkers returned home by each player.
        std::vector<std::vector<int>> board_;
        std::vector<int> home_;  // Checkers for each player.
        std::vector<TurnHistoryInfo> turn_history_info_;  // Info needed for Undo.

        void ApplyNormalAction(Action action);
        void ApplyChanceNodeAction(Action chance_outcome);

        std::set<int> GetBlockedPoints(std::vector<int> curr_board, std::vector<int> opp_board) const;
        std::set<int> GetHittablePoints(std::vector<int> board) const;
        std::vector<Action> ProcessLegalMoves(const std::set<CheckerMove>& legal_moves) const;
        std::string CurrentPlayerHumanReadable() const;
        std::string DiceHumanReadable() const;
        void InitalizeBoard();
        std::string GetBoardString() const;
        std::string GetXPlayerHomeString() const;
        std::string GetOPlayerHomeString() const;
};

class NannonGame : public Game {
    public:
        explicit NannonGame(const GameParameters& params);

        int NumDistinctActions() const override { return n_distinct_actions_; }
        std::unique_ptr<State> NewInitialState() const override {
          return std::unique_ptr<State>(
              new NannonState(shared_from_this(),
                                 n_points_,
                                 n_chex_per_player_,
                                 n_die_,
                                 chance_outcomes_,
                                 chance_outcome_values_,
                                 state_encoding_size_,
                                 home_pos_,
                                 score_pos_));
        }

        int MaxChanceOutcomes() const override { return n_die_; }

        // Arbitrarily chosen number to ensure the game is finite.
        int MaxGameLength() const override { return kMaxGameLength; }

        int NumPlayers() const override { return kNumPlayers; }
        double MinUtility() const override { return -MaxUtility(); }
        double UtilitySum() const override { return 0; } // Game is zero sum.
        double MaxUtility() const override;
        std::shared_ptr<const Game> Clone() const override {
          return std::shared_ptr<const Game>(new NannonGame(*this));
        }

        std::vector<int> ObservationTensorShape() const override {
          // Encode each point on the board as one double:
          // - One double for whether there is one checker or not (1 or 0).
          //
          // Return a vector encoding:
          // Every point listed for the current player.
          // Every point listed for the opponent.
          // One double for the number of checkers at home for the current player.
          // One double for the number of checkers scored for the current player.
          // One double for whether it's the current player's turn (1 or 0).
          // One double for the number of checkers at home for the opponent.
          // One double for the number of checkers scored for the opponent.
          // One double for whether it's the opponent's turn (1 or 0).

          return {state_encoding_size_};
        }
        int n_points() const { return n_points_;}
        int n_chex() const { return n_chex_per_player_;}
        int n_die() const { return n_die_;}

    private:
        const int n_points_;
        const int n_chex_per_player_;
        const int n_die_;
        const int n_distinct_actions_;
        const std::vector<std::pair<Action, double>> chance_outcomes_;
        const std::vector<int> chance_outcome_values_;
        const int state_encoding_size_;
        const int home_pos_;
        const int score_pos_;
};

}  // namespace open_spiel

#endif  // THIRD_PARTY_OPEN_SPIEL_GAMES_NANNON_H_
